UNIT List2ADT;
(* Unit List2ADT implements the List ADT in a linked structure.  *)
(*   The ordering is determined by applying function Compare to  *)
(*   two list items.                                             *)
(* Any program that uses this unit must provide unit List2Data   *)
(* which contains the following definitions:                     *)
(*   ListElementType - the type declaration for the list items   *)
(*   RelationType - (Less, Greater, Equal)                       *)
(*   Function Compare                                            *)
(*     (Item1, Item2 : ListElementType) : RelationType;          *)
(*       Less   : Item1 "comes before" Item2                     *)
(*       Greater: Item1 "comes after" Item2                      *)
(*       Equal  : Item1 and Item2 are the same                   *)

INTERFACE
(* Gain access to ListElementType, RelationType, and *)
(* function Compare.                                 *)
USES List2Data;
TYPE
  PointerType = ^NodeType;
  NodeType = RECORD
    Info   : ListElementType;               (* list user's data *)
    Next   : PointerType                (* pointer to next node *)
  END; (* NodeType *)

  ListType = RECORD
    Length       : Integer;       (* number of items in ListData *)
    CurrentItem  : PointerType;       (* pointer to current item *)
    ListData     : PointerType (* external pointer to list items *)
  END;  (* ListType *)


(*************************************************************)

PROCEDURE CreateList 
  (VAR List : ListType);
  (* Function: Initializes List to empty state. *)
  (* Pre:  List has not been created.           *)
  (* Post: List exists and is empty.            *)

(*************************************************************)

PROCEDURE DestroyList 
  (VAR List : ListType);
  (* Function: Destroys all elements, leaving List *)
  (*   in empty state.                             *)
  (* Pre:  List has been created.                  *)
  (* Post: List is empty.                          *)

(*************************************************************)

FUNCTION EmptyList
  (List : ListType ) : Boolean;
  (* Function: Determines whether List is empty. *)
  (* Pre:  List has been created.                *)
  (* Post: EmptyList = (List is empty)           *)

(*************************************************************)

FUNCTION FullList
  (List : ListType) : Boolean;
  (* Function: Determines whether List is full. *)
  (* Pre:  List has been created.               *)
  (* Post: FullList = (List is full)            *)

(*************************************************************)

PROCEDURE RetrieveElement
  (List : ListType;
   VAR Element  : ListElementType);
  (* Function: If Compare(Element, SomeElement) = Equal for some *)
  (*   component in List, Found is True and Element =SomeElement.*)
  (* Pre:  List has been created.                                *)
  (*       Key field(s) of Element are initialized.              *)
  (* Post: Found = (Compare(Element, SomeElement) = Equal)       *)
  (*       IF Found, Element = copy of SomeElement; otherwise,   *) 
  (*         Element is undefined except for key.                *)
  (*       List is unchanged.

(*************************************************************)

FUNCTION ListLength
  (List : ListType) : Integer;
  (* Function: Determines the number of elements in List. *)
  (* Pre:  List has been created.                         *)
  (* Post: Length = number of elements in List.           *)

(*************************************************************)

PROCEDURE InsertElement 
  (VAR List   : ListType;
   NewElement : ListElementType);
  (* Function: Adds NewElement to List.         *)
  (* Pre:  List has been created;               *)
  (*       List is not full;                    *)
  (*       NewElement is not in List;           *)
  (*       List is ordered by function Compare. *)
  (* Post: List = original list + NewElement;   *)
  (*       List is ordered by function Compare. *)

(*************************************************************)

PROCEDURE DeleteElement 
  (VAR List : ListType;
   Element  : ListElementType);
  (* Function: Deletes the element where Compare(Element, *)
  (*   SomeElement) = Equal.                              *)
  (* Pre:  List has been created.                         *)
  (*       Key field(s) of Element are initialized.       *)
  (*       Compare(Element, SomeElement) = Equal for one  *)
  (*         and only one element in List.                *)
  (* Post: List = original list with SomeElement removed  *)
  (*         where Compare(Element, SomeElement) = Equal. *)
  (*         List is still ordered by Compare.            *)

(*************************************************************)

PROCEDURE ResetList 
  (VAR List : ListType);
  (* Function: Initializes current element to predecessor *)
  (*   of first element in list.                          *)
  (* Pre:  List has been created.                         *)
  (* Post: Successor of current element is first element. *)

(*************************************************************)

PROCEDURE GetNextItem 
  (VAR List : ListType;
   VAR Element  : ListElementType);
  (* Function: Get the successor of current element.      *)

  (* Pre:  List has been created.                         *)
  (*       Current element has been defined.              *)
  (* Post: Current element = successor of current element *)
  (*       Element = current element                      *)

(*************************************************************)
(*-----------------------------------------------------------*)
(*************************************************************)

IMPLEMENTATION

PROCEDURE CreateList
  (VAR List : ListType);
  (* Pre:  List has not been created.  *)
  (* Post: List.ListData is set to NIL. *)
BEGIN (* CreateList *)
  List.ListData := NIL;
  List.Length := 0
END; (* CreateList *) 

(*************************************************************)

FUNCTION EmptyList
  (List : ListType) : Boolean;
  (* Pre:  List has been created.            *)
  (* Post: EmptyList = (List.ListData = NIL) *)

BEGIN (* EmptyList *)
  EmptyList := (List.ListData =  NIL)
END;  (* EmptyList *)

(*************************************************************)

FUNCTION FullList
  (List : ListType) : Boolean;
  (* Post:  FullList = False *)
BEGIN (* FullList *)
  FullList := False
END;  (* FullList *)

(*************************************************************)

FUNCTION ListLength
  (List : ListType) : Integer;
  (* Pre:  List has been initialized. *)
  (* Post: Length is List.Length.     *)
BEGIN  (* Length *)
  ListLength := List.Length
END;   (* Length *)

(*************************************************************)

PROCEDURE DestroyList
  (VAR List : ListType);
  (* Pre:  List has been created *)
  (* Post: List.Length = 0; List.ListData = NIL            *)
  (*       All items in List.ListData have been disposed.  *)
VAR
  TempPtr : PointerType;

BEGIN (* DestroyList *)
  WITH List DO
    BEGIN
      List.Length := 0;
      (* Remove all elements from the list, freeing the space. *)
      WHILE ListData <>  NIL DO
        BEGIN
          (* Unlink the first node from the list. *)
          TempPtr := ListData;
          ListData := ListData^.Next;
          (* Free the space from the node. *)
          Dispose (TempPtr)
        END (* WHILE *)
      END  (* WITH *)

END; (* DestroyList *)

(*************************************************************)

PROCEDURE RetrieveElement
  (List        : ListType;
   VAR Element : ListElementType);

VAR
  Location     : PointerType;

BEGIN  (* RetrieveElement *)
  Location := List.ListData;
  WHILE Location^.Info.Position <> Element.Position DO
    Location := Location^.Next;
  Element := Location^.Info;
END;   (* RetrieveElement *)

(*************************************************************)

PROCEDURE InsertElement
  (VAR List   : ListType;
   NewElement : ListElementType);
  (* Pre:  List has been initialized.*)
  (*       Compare (NewElement, SomeElement) <> Equal *)
  (* Post: Node with Info field of NewElement is in   *)
  (*       List;  List is ordered by Compare.         *)

VAR
  NewNode      : PointerType;  (* pointer to the new node   *)
  PredLoc      : PointerType;  (* pointer to preceding node *)
  Location     : PointerType;    (* used for searching loop *)
  MoreToSearch : Boolean;

BEGIN  (* InsertElement *)
  WITH List DO
    BEGIN 
      (* initialize loop variables *)
      Location := List.ListData;
      PredLoc := NIL;
      MoreToSearch := Location <> NIL;
      List.Length := List.Length + 1;

      (* Find insertion point. *)
      WHILE MoreToSearch DO
        CASE CompareName (NewElement, Location^.Info) OF
          Greater : BEGIN
                      PredLoc := Location;
                      Location := Location^.Next;
                      MoreToSearch := Location <> NIL
                    END;
          Less    : MoreToSearch := False
        END;  (* case *)

      (* Create space for new element. *)
      New(NewNode);
      NewNode^.Info := NewElement;

      (* Insert the new node into the list. *)
      IF PredLoc = NIL
        THEN (* list is empty OR inserting at beginning of list *)
          BEGIN
            NewNode^.Next := ListData;
            ListData := NewNode
          END (* inserting first element *)

        ELSE (* inserting in the middle or at the end *)
          BEGIN
            NewNode^.Next := PredLoc^.Next;
            PredLoc^.Next := NewNode
          END (* inserting in middle or at end *)
    END (* WITH *)
END;  (* InsertElement *)

(*************************************************************)

PROCEDURE DeleteElement 
  (VAR List   : ListType;
   Element    : ListElementType);
  (* Pre:  List has been initialized.                  *)
  (*       Compare (Element, SomeElement) = Equal      *)
  (* Post: Node with Info field SomeElement where      *)
  (*         Compare (Element, SomeElement) = Equal    *)
  (*         is removed from the ListData and disposed.*)
  (*       List is ordered by Compare.                 *)

VAR
  PredLoc   : PointerType;  (* pointer to preceding node *)
  Location  : PointerType;    (* used for searching loop *)

  TempPtr   : PointerType;         (* used for disposing *)
  MoreToSearch : Boolean;
BEGIN  (* DeleteElement *)
  WITH List DO
    BEGIN 
      (* initialize loop variables *)
      Location := ListData;
      PredLoc := NIL;
      MoreToSearch := Location <> NIL;
      Length := Length - 1;
      (* Find insertion point. *)
      WHILE MoreToSearch DO
        CASE CompareName (Element, Location^.Info) OF
          Greater: BEGIN
                     PredLoc := Location;
                     Location := Location^.Next;
                   END;
          Equal  : MoreToSearch := False
        END;  (* case *)

      (* delete Location *)
      IF PredLoc = NIL
        THEN (* deleting first element *)
          BEGIN
            TempPtr := Location;
            ListData := Location^.Next;
          END 

        ELSE (* deleting middle or last element *)
          BEGIN
            TempPtr := Location;
            PredLoc^.Next := Location^.Next
          END;
      Dispose (TempPtr)
    END (* WITH *)
END;  (* DeleteElement *)

(*************************************************************)

PROCEDURE ResetList
  (VAR List : ListType);
  (* Pre:  List has been created.                    *)
  (* Post: List.CurrentItem is set to List.ListData. *)
BEGIN  (* ResetList *)
  List.CurrentItem := List.ListData
END;   (* ResetList *)

(***********************************************************)

PROCEDURE GetNextItem
  (VAR List        : ListType;
   VAR Element : ListElementType);
  (* Pre:  List has been created.                    *)
  (*       List.CurrentItem has been initialized.    *)
  (*       GetNextItem is not called when            *)
  (*         List.CurrentItem is NIL.                *)
  (* Post: Element is set to List.CurrentItem^.Info. *)
  (*       CurrentItem is set to CurrentItem^.Next.  *)
BEGIN  (* GetNextItem *)
  Element := List.CurrentItem^.Info;
  List.CurrentItem := List.CurrentItem^.Next
END;   (* GetNextItem *)

END.




